import reapy
import reapy.reascript_api as RPR
import contextlib
from .defer import ReaperConsole

import collections
import io
import os
import sys
import typing as ty

_ORIGINAL_PRINT = print


def add_reascript(path: str, section_id: int = 0, commit: bool = True) -> int:
    """
    Add a ReaScript and return the new action ID.

    Parameters
    ----------
    path : str
        Path to script.
    section_id : int, optional (default=0, corresponds to main section).
        Action section ID to which the script must be added.
    commit : bool, optional
        Whether to commit change. Use it when adding a single script.
        You can optimize bulk adding `n` scripts by setting
        `commit=False` for the first `n-1` calls and `commit=True` for
        the last call.

    Returns
    -------
    action_id : int
        New ReaScript action ID.
    """
    ...


def arm_command(command_id: int, section: str = "") -> None:
    """
    Arm or disarm command.

    Parameters
    ----------
    command_id : int
        Command ID. If 0, disarm command.
    section : str, optional
        Command section. Empty string for main section. Default="".
    """
    ...


def browse_for_file(window_title: str = "",
                    extension: str = "") -> ty.Optional[str]:
    """
    Ask the user to select a file.

    Parameters
    ----------
    window_title : str, optional
        Window title (default="")
    extension : str, optional
        Extension for file (e.g. "mp3", "txt"...) (default=all types).

    Returns
    -------
    path : str or NoneType
        Path to file, or None if user cancelled.
    """
    ...


def clear_console() -> None:
    """
    Clear Reaper console.

    See also
    --------
    ReaProject.show_console_message
    """
    ...


def clear_peak_cache() -> None:
    """
    Reset global peak cache.
    """
    ...


def dB_to_slider(db: float) -> float:
    """
    Convert decibel value to slider.

    Parameters
    ----------
    db : float
        Decibel value.

    Returns
    -------
    slider : float
        Slider value.

    See also
    --------
    slider_to_dB
    """
    ...


def delete_ext_state(section: str, key: str, persist: bool = False) -> None:
    """
    Delete extended state value for a given section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key.
    persist : bool
        Whether extended state should remain deleted next time REAPER
        is opened.
    """
    ...


def disarm_command() -> None:
    """
    Disarm command.
    """
    ...


def get_armed_command() -> ty.Optional[ty.Tuple[int, str]]:
    ...


def get_command_id(command_name: str) -> ty.Optional[int]:
    """
    Return ID of command with a given name.

    Parameters
    ----------
    command_name : str
        Command name.

    Returns
    -------
    command_id : int or None
        Command ID, or None if name can't be found.
    """
    ...


def get_command_name(command_id: int) -> ty.Optional[str]:
    """
    Return name of command with a given ID.

    Parameters
    ----------
    command_id : int
        Command ID.

    Returns
    -------
    command_name : str, None
        Command name, or None for a native command.
    """
    ...


def get_exe_dir() -> str:
    """
    Return REAPER.exe directory (e.g. "C:\\Program Files\\REAPER").

    Returns
    -------
    path : str
        Path to REAPER.exe directory.
    """
    ...


def get_ext_state(section: str, key: str) -> str:
    """
    Get the extended state value for a specific section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key for section `section`.

    Returns
    -------
    value : str
        Extended state value.

    See also
    --------
    delete_ext_state
    set_ext_state
    """
    ...


def get_global_automation_mode() -> str:
    """
    Return global automation override mode.

    Returns
    -------
    override_mode : str
        One of the following values:
            "bypass"
            "latch"
            "none"
            "read"
            "touch"
            "trim/read"
            "write"
    """
    ...


def get_ini_file() -> str:
    """
    Return path to REAPER.ini file.

    Returns
    -------
    path : str
        Path to REAPER.ini file.
    """
    ...


def get_last_touched_track() -> ty.Optional[reapy.Track]:
    """
    Return last touched track, or None if no track has been touched.

    Returns
    -------
    track : Track or None if no track has been touched.
    """
    ...


def get_main_window() -> reapy.Window:
    """
    Return main window.

    Returns
    -------
    window : Window
        Main window.
    """
    ...


@reapy.inside_reaper()
def get_projects() -> ty.List[reapy.Project]:
    """
    Return list of all opened projects.

    Returns
    -------
    projects : list of Project
        List of all projects.
    """
    ...


def get_reaper_version() -> str:
    ...


def get_resource_path() -> str:
    """
    Return path to directory where .ini files are stored.

    Returns
    -------
    path : str
        Path to directory where .ini files are stored.
    """
    ...


def get_user_inputs(
    title: str,
    captions: ty.List[str],
    retvals_size: int = 1024
) -> ty.Dict[str, str]:
    """Show text inputs to user and get values from them.

    Parameters
    ----------
    title : str
        Popup title.
    captions : List[str]
        Names of input fields.
    retvals_size : int, optional
        Maximum number of characters that will be retrieved for each
        field. User may enter more, but only the first `retvals_size`
        will be returned. (default=1024)

    Returns
    -------
    Dict[str,str]
        Dictionary of pairs {caption: response}.

    Raises
    ------
    RuntimeError
        When user clicked the Cancel button.
    """
    ...


def has_ext_state(section: str, key: str) -> bool:
    """
    Return whether extended state exists for given section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key.

    Returns
    -------
    has_ext_state : bool
    """
    ...


def add_project_tab(make_current_project: bool = True) -> reapy.Project:
    """Open new project tab."""
    ...


def open_project(
    filepath: str, in_new_tab: bool = False, make_current_project: bool = True
) -> reapy.Project:
    """
    Open project and return it.

    Parameters
    ----------
    filepath : str
    in_new_tab : bool, optional
        Focus will be on new tab.

    Returns
    -------
    project : Project
        Opened project.
    """
    ...


def perform_action(action_id: int) -> None:
    """
    Perform action with ID `action_id` in the main Actions section.

    Parameters
    ----------
    action_id : int
        Action ID in the main Actions section.
    """
    ...


class prevent_ui_refresh(contextlib.ContextDecorator):
    """Class to prevent UI refresh on certain pieces of code.

    Its instance can be used both as decorator and as context manager:

    >>> with reapy.prevent_ui_refresh():
    ...     reapy.Project.add_track()

    >>> @prevent_ui_refresh()
    >>> def some_function(*args, **kwargs):
    ...     reapy.Project.add_track()

    """

    def __enter__(self) -> None:
        ...

    def __exit__(self, exc_type: Exception, exc_val: ty.Any,
                 exc_tb: ty.Any) -> None:
        ...


def print(*args: ty.Any, **kwargs: ty.Any) -> None:
    """
    Alias to ReaProject.show_console_message.
    """
    ...


class reaprint(contextlib.ContextDecorator):
    """Class to send all prints to ReaperConsole.

    Its instance can be used both as decorator and context manager:

    >>> with reapy.reaprint():
    ...     print('This will go to the console!')
    ...     print('All these contexted will go to the console!')

    >>> @reapy.reaprint()
    >>> def some_function(*args, **kwargs):
    ...     print('This will go to the console!')
    ...     print('All these decorated prints will go to the console!')

    """
    _original_stdouts: ty.Deque[str]

    def __enter__(self) -> None:
        ...

    def __exit__(self, exc_type: ty.Any, exc_val: ty.Any,
                 exc_tb: ty.Any) -> None:
        ...


def remove_reascript(path: str, section_id: int = 0,
                     commit: bool = True) -> None:
    """
    Remove a ReaScript.

    Parameters
    ----------
    path : str
        Path to script.
    section_id : int, optional (default=0, corresponds to main section).
        Action section ID to which the script must be added.
    commit : bool, optional
        Whether to commit change. Use it when removing a single script.
        You can optimize bulk removing `n` scripts by setting
        `commit=False` for the first `n-1` calls and `commit=True` for
        the last call.
    """
    ...


def rgb_from_native(native_color: int) -> ty.Tuple[int, int, int]:
    """
    Extract RGB values from a native (OS-dependent) color.

    Parameters
    ----------
    native_color : int
        Native color.

    Returns
    -------
    r, g, b : (int, int, int)
        RGB values between 0 and 255.
    """
    ...


def rgb_to_native(rgb: ty.Tuple[int, int, int]) -> int:
    """
    Make a native (OS-dependent) color from RGB values.

    Parameters
    ----------
    rgb : (int, int, int)
        RGB triplet of integers between 0 and 255.

    Returns
    -------
    native_color : int
        Native color.
    """
    ...


def set_ext_state(section: str, key: str, value: str,
                  persist: bool = False) -> None:
    """
    Set the extended state value for a specific section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key for section `section`.
    value : str
        Extended state value for section `section` and key `key`.
    persist : bool
        Whether the value should be stored and reloaded the next time
        REAPER is opened.

    See also
    --------
    delete_ext_state
    get_ext_state
    """
    ...


def set_global_automation_mode(mode: str) -> None:
    """
    Set global automation mode.

    Parameters
    ----------
    mode : str
        One of the following values:
            "bypass"
            "latch"
            "none"
            "read"
            "touch"
            "trim/read"
            "write"
    """
    ...


def show_console_message(*args: ty.Any, sep: str = " ",
                         end: str = "\n") -> None:
    """
    Print a message to the Reaper console.

    Parameters
    ----------
    args : tuple
        Values to print.
    sep : str, optional
        String inserted between values (default=" ").
    end : str, optional
        String appended after the last value (default="\n").
    """
    ...


def show_message_box(text: str = "", title: str = "", type: str = "ok") -> str:
    """
    Show message box.

    Parameters
    ----------
    text : str
        Box message
    title : str
        Box title
    type : str
        One of the following values.

        "ok"
        "ok-cancel"
        "abort-retry-ignore"
        "yes-no-cancel"
        "yes-no"
        "retry-cancel"

    Returns
    -------
    status : str
        One of the following values.

        "ok"
        "cancel"
        "abort"
        "retry"
        "ignore"
        "yes"
        "no"
    """
    ...


def slider_to_dB(slider: float) -> float:
    """
    Convert slider value to decibel.

    Parameters
    ----------
    slider : float
        Slider value.

    Returns
    -------
    db : float
        Decibel value.

    See also
    --------
    dB_to_slider
    """
    ...


def test_api() -> None:
    """Display a message window if the API can successfully be called."""
    ...


class undo_block(contextlib.ContextDecorator):
    """Class to register undo block.

    Its instance can be used both as decorator and context manager:

    >>> with reapy.undo_block('add track'):
    ...     reapy.Project.add_track()

    >>> @reapy.undo_block('add track')
    >>> def some_function(*args, **kwargs):
    ...     reapy.Project.add_track()

    :param undo_name: Str to register undo name (shown later in Undo menu)
    :param flags: Int to pass to Undo_EndBlock
                    (leave default if you don't know what it is)
    """
    undo_name: str
    flags: int

    def __init__(self, undo_name: str, flags: int = 0):
        ...

    def __enter__(self) -> None:
        ...

    def __exit__(self, exc_type: ty.Any, exc_val: ty.Any,
                 exc_tb: ty.Any) -> None:
        ...


def update_arrange() -> None:
    """
    Redraw the arrange view.
    """
    ...


def update_timeline() -> None:
    """
    Redraw the arrange view and ruler.
    """
    ...


def view_prefs() -> None:
    """
    Open Preferences.
    """
    ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Envelope(ReapyObject):
    id: int
    _parent: ty.Union[ty.Type[reapy.Take], ty.Type[reapy.Track]]

    def __init__(self, parent: ReapyObject, id: int) -> None:
        ...

    @property
    def _args(self) -> ty.Tuple[ReapyObject, int]:
        ...

    def add_item(self, position: float = 0., length: float = 1.,
                 pool: int = 0) -> reapy.AutomationItem:
        """
        Add automation item to envelope.

        Parameters
        ----------
        position : float, optional
            New item position in seconds (default=0).
        length : float
            New item length in seconds (default=0).
        pool : int
            New item pool index. If >= 0 the automation item will be a
            new instance of that pool (which will be created as an
            empty instance if it does not exist).

        Returns
        -------
        item : reapy.AutomationItem
            New automation item.
        """
        ...

    def delete_points_in_range(self, start: float, end: float) -> None:
        """
        Delete envelope points between `start` and `end`.

        Parameters
        ----------
        start : float
            Range start in seconds.
        end : float
            Range end in seconds.
        """
        ...

    @reapy.inside_reaper()
    def get_derivatives(self, time: float,
                        raw: bool = False) -> ty.Tuple[float, float, float]:
        """
        Return envelope derivatives of order 1, 2, 3 at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw values or the human-readable version
            which is printed in REAPER GUI (default=False).

        Returns
        -------
        d, d2, d3 : float
            First, second and third order derivatives.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_derivatives(10, raw=True)
        (0.10635556358181712, 0.2127113398749741, 0.21271155258652666)
        >>> envelope.get_value(10)  # human-readable
        ('10%L', '21%L', '21%L')
        """
        ...

    @reapy.inside_reaper()
    def get_value(self, time: float,
                  raw: bool = False) -> ty.Union[float, str]:
        """
        Return envelope value at a given time.

        Parameters
        ----------
        time : float
            Time in seconds.
        raw : bool, optional
            Whether to return raw value or its human-readable version,
            which is the one that is printed in REAPER GUI
            (default=False).

        Returns
        -------
        value : float or str
            Envelope value.

        Examples
        --------
        >>> envelope = track.envelopes["Pan"]
        >>> envelope.get_value(10, raw=True)
        -0.5145481809245827
        >>> envelope.get_value(10)  # human-readable
        '51%R'
        """
        ...

    @property
    def items(self) -> ty.List[reapy.AutomationItem]:
        """
        List of automation items in envelope.

        :type: list of reapy.AutomationItem
        """
        ...

    @property
    def n_items(self) -> int:
        """
        Number of automation items in envelope.

        :type: int
        """
        ...

    @property
    def n_points(self) -> int:
        """
        Number of points in envelope.

        :type: int
        """
        ...

    @property
    def name(self) -> str:
        """
        Envelope name.

        :type: str
        """
        ...

    @property
    def parent(self) -> ty.Union[ty.Type[reapy.Take], ty.Type[reapy.Track]]:
        """
        Envelope parent.

        :type: Take or Track
        """
        ...


class EnvelopeList(ReapyObject):
    """
    Container class for the list of envelopes on a Take or Track.

    Envelopes can be accessed from the EnvelopeList either by index,
    name or chunk_name (e.g. "<VOLENV").

    Examples
    --------
    >>> len(track.envelopes)
    2
    >>> envelope = track.envelopes[0]
    >>> envelope.name
    'Volume'
    >>> envelope == track.envelopes["Volume"]
    True
    >>> envelope == track.envelopes["<VOLENV"]
    True
    >>> [e.name for e in track.envelopes]
    ['Volume', 'Pan']
    """
    parent: ty.Union[ty.Type[reapy.Take], ty.Type[reapy.Track]]

    def __init__(self,
                 parent: ty.Union[ty.Type[reapy.Take], ty.Type[reapy.Track]]
                 ) -> None:
        ...

    @property
    def _args(
            self
    ) -> ty.Tuple[ty.Union[ty.Type[reapy.Take], ty.Type[reapy.Track]]]:
        ...

    def __getitem__(self, key: ty.Union[str, int]) -> Envelope:
        ...

    def __len__(self) -> int:
        ...



from reapy.errors import DisconnectedClientError, DistError
from reapy.tools import json
from .socket import Socket


class Client(Socket):

    def __init__(self, port, host="localhost"):
        super().__init__()
        self._connect(port, host)
        self.port, self.host = port, host

    def _connect(self, port, host):
        super().connect((host, port))
        self.address = self.recv(timeout=None).decode("ascii")

    def _get_result(self):
        s = self.recv(timeout=None).decode()
        return json.loads(s)

    def request(self, function, input=None):
        request = {"function": function, "input": input}
        request = json.dumps(request).encode()
        self.send(request)
        result = self._get_result()
        if result["type"] == "result":
            return result["value"]
        elif result["type"] == "error":
            raise DistError(result["traceback"])



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class AudioAccessor(ReapyObject):
    id: bytes

    def __init__(self, id: bytes) -> None:
        ...

    @property
    def _args(self) -> ty.Tuple[bytes]:
        ...

    def delete(self) -> None:
        """Delete audio accessor."""
        ...

    @property
    def end_time(self) -> float:
        """
        End time of audio that can be returned from this accessor.

        Return value is in seconds.

        :type: float
        """
        ...

    def get_samples(self,
                    start: float,
                    n_samples_per_channel: int,
                    n_channels: int = 1,
                    sample_rate: int = 44100) -> ty.List[float]:
        """
        Return audio samples.

        Parameters
        ----------
        start : float
            Start time in seconds.
        n_samples_per_channel : int
            Number of required samples per channel
        n_channels : int, optional
            Number of required channels (default=1).
        sample_rate : float, optional
            Required sample rate (default=44100).

        Returns
        -------
        samples : list
            List of length n_samples*n_channels.

        Examples
        --------
        To separate channels use:

        >>> samples = audio_accessor.get_samples(0, 1024, 2)
        >>> first_channel = samples[::2]
        >>> second_channel = samples[1::2]
        """
        ...

    @property
    def has_state_changed(self) -> bool:
        """
        Whether underlying state has changed.

        :type: bool
        """
        ...

    def hash(self) -> str:
        """
        String that changes only if the underlying samples change.

        :type: str
        """
        ...

    @property
    def start_time(self) -> float:
        """
        Start time of audio that can be returned from this accessor.

        Return value is in seconds.

        :type: float
        """
        ...



import sys


def is_inside_reaper():
    """
    Return whether ``reapy`` is imported from inside REAPER.

    If ``reapy`` is run from inside a REAPER instance but currently
    controls another REAPER instance on a slave machine (with
    ``reapy.connect``), return False.
    """
    inside = hasattr(sys.modules["__main__"], "obj")
    if not inside:
        return False
    else:
        try:
            return machines.get_selected_machine_host() is None
        except NameError:
            # machines is undefined because we are still in the initial
            # import process.
            return True



from .tools import (
    connect, connect_to_default_machine, dist_api_is_enabled, inside_reaper,
    reconnect
)
from . import reascript_api
from .config import configure_reaper
from .core import *
from .core.reaper import *


# This is a fork of python-reapy maintained by Elie Gavoty
__version__ = "0.10.0-renardo"



"""Define tools such as ``inside_reaper`` and custom json module."""

import reapy
from ._inside_reaper import inside_reaper, dist_api_is_enabled
from .network.machines import connect, connect_to_default_machine, reconnect
from .extension_dependency import depends_on_sws, depends_on_extension



"""Define FX and FXParam classes."""

import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
from reapy.errors import DistError, UndefinedFXParamError
import typing as ty


class FX(ReapyObject):
    """FX on a Track or a Take."""

    _class_name: str
    functions: ty.Dict[str, ty.Dict[str, ty.Callable[..., ty.Any]]]
    parent_id: str
    index: int

    def __init__(self,
                 parent: ty.Optional[ty.Union[reapy.Track, reapy.Take]] = None,
                 index: ty.Optional[int] = None,
                 parent_id: ty.Optional[str] = None) -> None:
        ...

    def _get_functions(self) -> ty.Dict[str, ty.Callable[..., ty.Any]]:
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, ty.Union[str, int]]:
        ...

    def close_chain(self) -> None:
        """Close FX chain."""
        ...

    def close_floating_window(self) -> None:
        """Close FX floating window."""
        ...

    def close_ui(self) -> None:
        """Close user interface."""
        ...

    def copy_to_take(self, take: reapy.Take, index: int = 0) -> None:
        """
        Copy FX to take.

        Parameters
        ----------
        take : Take
            Destination take.
        index : int
            Index on destination take.

        See also
        --------
        FX.move_to_take
        """
        ...

    def copy_to_track(self, track: reapy.Track, index: int = 0) -> None:
        """
        Copy FX to track.

        Parameters
        ----------
        track : Track
            Destination track.
        index : int
            Index on destination track.

        See also
        --------
        FX.move_to_track
        """
        ...

    def delete(self) -> None:
        """Delete FX."""
        ...

    def disable(self) -> None:
        """Disable FX."""
        ...

    def enable(self) -> None:
        """Enable FX."""
        ...

    @property
    def is_enabled(self) -> bool:
        """
        Whether FX is enabled.

        :type: bool
        """
        ...

    @is_enabled.setter
    def is_enabled(self, enabled: bool) -> None:
        ...

    @property
    def is_online(self) -> bool:
        """
        Whether FX is online.

        :type: bool
        """
        ...

    @is_online.setter
    def is_online(self, online: bool) -> None:
        ...

    @property
    def is_ui_open(self) -> bool:
        """
        Whether FX user interface is open.

        :type: bool
        """
        ...

    @is_ui_open.setter
    def is_ui_open(self, open: bool) -> None:
        ...

    def make_offline(self) -> None:
        """Make FX offline."""
        ...

    def make_online(self) -> None:
        """Make FX online."""
        ...

    def move_to_take(self, take: reapy.Take, index: int = 0) -> None:
        """
        Move FX to take.

        Parameters
        ----------
        take : Take
            Destination take.
        index : int
            Index on destination take.

        See also
        --------
        FX.copy_to_take
        """
        ...

    def move_to_track(self, track: reapy.Track, index: int = 0) -> None:
        """
        Move FX to track.

        Parameters
        ----------
        track : Track
            Destination track.
        index : int
            Index on destination track.

        See also
        --------
        FX.copy_to_track
        """
        ...

    @property
    def n_inputs(self) -> int:
        """
        Number of inputs of FX.

        :type: int
        """
        ...

    @property
    def n_outputs(self) -> int:
        """
        Number of outputs of FX.

        :type: int
        """
        ...

    @property
    def n_params(self) -> int:
        """
        Number of parameters.

        :type: int
        """
        ...

    @property
    def n_presets(self) -> int:
        """
        Number of presets.

        :type: int
        """
        ...

    @property
    def name(self) -> str:
        """
        FX name.

        :type: str
        """
        ...

    def open_chain(self) -> None:
        """Open FX chain with focus on FX."""
        ...

    def open_floating_window(self) -> None:
        """Open FX floating window."""
        ...

    def open_ui(self) -> None:
        """Open FX user interface."""
        ...

    @property
    def params(self) -> reapy.FXParamsList:
        """
        List of parameters.

        :type: FXParamsList
        """
        ...

    @property
    def parent(self) -> ty.Union[reapy.Track, reapy.Take]:
        """
        FX parent.

        :type: Track or Take
        """
        ...

    @property
    def preset(self) -> str:
        """
        FX preset name.

        :type: str

        Attribute can be set by passing a str or int. In the first
        case, the str can either be a preset name or the path to a
        .vstpreset file. Otherwise, the int is the preset index.
        """
        ...

    @preset.setter
    def preset(self, preset: ty.Union[str, int]) -> None:
        """
        Set FX preset.

        Parameters
        ----------
        preset : str or int
            If str, preset name or path to .vstpreset file. If int,
            preset index. Set to -2 for factory preset, and -1 for user
            default preset.
        """
        ...

    @property
    def preset_index(self) -> int:
        """
        FX preset index.

        :type: int
        """
        ...

    @property
    def preset_file(self) -> str:
        """
        Path to FX preset file.

        :type: str
        """
        ...

    def use_previous_preset(self) -> None:
        """Use previous preset in the presets list."""
        ...

    def use_next_preset(self) -> None:
        """Use next preset in the presets list."""
        ...

    @property
    def window(self) -> ty.Optional[reapy.Window]:
        """
        Floating window associated to FX, if it exists.

        :type: Window or NoneType
        """
        ...


class FXList(ReapyObjectList):
    """
    Container class for a list of FXs.

    FXs can be accessed by name or index.

    Examples
    --------
    >>> fx_list = track.fxs
    >>> fx_list[0]
    FX(parent_id="(MediaTrack*)0x0000000006CDEBE0", index=0)
    >>> len(fx_list)
    1
    >>> fx_list["VST: ReaComp (Cockos)"]
    FX(parent_id="(MediaTrack*)0x0000000006CDEBE0", index=0)
    """

    _class_name = "FXList"
    parent: ty.Union[reapy.Track, reapy.Take]

    def __init__(self, parent: ty.Union[reapy.Track, reapy.Take]) -> None:
        ...

    @reapy.inside_reaper()
    def __delitem__(self, key: ty.Union[int, slice]) -> None:
        ...

    @ty.overload
    def __getitem__(self, i: ty.Union[int, str]) -> reapy.FX:
        ...

    @ty.overload
    def __getitem__(self, i: slice) -> ty.List[reapy.FX]:
        ...

    def __len__(self) -> int:
        ...

    @reapy.inside_reaper()
    def _get_items_from_slice(self, slice: slice) -> ty.List[reapy.FX]:
        ...

    def _get_fx_index(self, name: str) -> int:
        ...

    @property
    def _args(self) -> ty.Tuple[ty.Union[reapy.Track, reapy.Take]]:
        ...



"""
Activate ``renardo_reapy`` server.

Running this ReaScript from inside REAPER sets the ``renardo_reapy`` server
that receives and executes API calls requests from outside. It will
automatically be run when importing ``renardo_reapy`` from outside, if it is
enabled.
"""

import pathlib
import site
import sys

try:
    import reapy
except ImportError:
    reapy_path = pathlib.Path(sys.path[0]).resolve().parent.parent
    sys.path.append(str(reapy_path))
    import reapy

from reapy.tools.network import Server


def run_main_loop():
    # Get new connections
    SERVER.accept()
    # Process API call requests
    requests = SERVER.get_requests()
    results = SERVER.process_requests(requests)
    SERVER.send_results(results)
    # Run main_loop again
    reapy.defer(run_main_loop)


def get_new_renardo_reapy_server():
    server_port = reapy.config.REAPY_SERVER_PORT
    reapy.set_ext_state("renardo_reapy", "server_port", server_port)
    server = Server(server_port)
    return server


if __name__ == "__main__":
    SERVER = get_new_renardo_reapy_server()
    run_main_loop()
    reapy.at_exit(reapy.delete_ext_state, "renardo_reapy", "server_port")



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
from reapy.errors import UndefinedEnvelopeError
import typing as ty


class Track(ReapyObject):
    """
    REAPER Track.

    Parameters
    ----------
    id : str or int
        If str, can either be a ReaScript ID (usually looking like
        ``"(MediaTrack*)0x00000000110A1AD0"``), or a track name. In
        that case, ``project`` must be specified.
        If int, the index of the track. In that case, ``project`` must
        be specified.
    project : Project
        Parent project of the track. Only necessary to retrieve a
        track from its name or index.

    Examples
    --------
    In most cases, accessing tracks is better done directly from
    the parent Project:

    >>> project = reapy.Project()
    >>> project.tracks[0]
    Track("(MediaTrack*)0x00000000110A1AD0")
    >>> project.tracks["PIANO"]  # This is actually the same track
    Track("(MediaTrack*)0x00000000110A1AD0")

    But the same track can also directly be instantiated with:

    >>> reapy.Track(0, project)
    Track("(MediaTrack*)0x00000000110A1AD0")

    or

    >>> reapy.Track("PIANO")
    Track("(MediaTrack*)0x00000000110A1AD0")
    """
    id: str
    _project: reapy.Project

    def __init__(
        self,
        id: ty.Union[str, int, float],
        project: ty.Optional[reapy.Project] = None
    ) -> None:
        ...

    @property
    def _args(self) -> ty.Tuple[ty.Union[str, int]]:
        ...

    @classmethod
    def _get_id_from_pointer(cls, id_: ty.Union[int, float]) -> str:
        ...

    def _get_project(self) -> reapy.Project:
        """
        Return parent project of track.

        Should only be used internally; one should directly access
        Track.project instead of calling this method.
        """
        ...

    def add_audio_accessor(self) -> reapy.AudioAccessor:
        """
        Create audio accessor and return it.

        Returns
        -------
        audio_accessor : AudioAccessor
            Audio accessor on track.
        """
        ...

    def add_fx(
        self, name: str, input_fx: bool = False, even_if_exists: bool = True
    ) -> reapy.FX:
        """
        Add FX to track and return it.

        Parameters
        ----------
        name : str
            FX name.
        input_fx : bool, optional
            Whether the FX should be an input (aka recording) FX or a
            standard FX (default=False). Note that if the track is the
            master track, input_fx=True will create a monitoring FX.
        even_if_exists : bool, optional
            Whether the FX should be added even if there already is an
            instance of the same FX on the track (default=True).

        Returns
        -------
        fx : FX
            New FX on track (or previously existing instance of FX if
            even_if_exists=False).

        Raises
        ------
        ValueError
            If there is no FX with the specified name.
        """
        ...

    def add_item(
        self,
        start: float = 0,
        end: ty.Optional[float] = None,
        length: float = 0
    ) -> reapy.Item:
        """
        Create new item on track and return it.

        Parameters
        ----------
        start : float, optional
            New item start in seconds (default=0).
        end : float, optional
            New item end in seconds (default None). If None, `length`
            is used instead.
        length : float, optional
            New item length in seconds (default 0).

        Returns
        -------
        item : Item
            New item on track.
        """
        ...

    def add_midi_item(
        self, start: float = 0, end: float = 1, quantize: bool = False
    ) -> reapy.Item:
        """
        Add empty MIDI item to track and return it.

        Parameters
        ----------
        start : float, optional
            New item start in seconds (or beats if `quantize`=True).
        end : float, optional
            New item end in seconds (or beats if `quantize`=True).
        quantize : bool, optional
            Whether to count time in beats (True) or seconds (False,
            default).
        """
        ...

    def add_send(
        self, destination: ty.Optional[reapy.Track] = None
    ) -> reapy.Send:
        """
        Add send to track and return it.

        Parameters
        ----------
        destination : Track or None
            Send destination (default=None). If None, destination is
            set to hardware output.

        Returns
        -------
        send : Send
            New send on track.
        """
        ...

    @property
    def automation_mode(self) -> str:
        """
        Track automation mode.

        One of the following values:
            "latch"
            "latch preview"
            "read"
            "touch"
            "trim/read"
            "write"

        :type: str
        """
        ...

    @automation_mode.setter
    def automation_mode(self, mode: str) -> None:
        """
        Set track automation mode.

        Parameters
        -------
        mode : str
            One of the following values:
                "latch"
                "latch preview"
                "read"
                "touch"
                "trim/read"
                "write"
        """
        ...

    @property
    def color(self) -> ty.Tuple[int, int, int]:
        """
        Track color in RGB format.

        :type: tuple of int
        """
        ...

    @color.setter
    def color(self, color: ty.Tuple[int, int, int]) -> None:
        """
        Set track color to `color`

        Parameters
        ----------
        color : tuple
            Triplet of integers between 0 and 255 corresponding to RGB
            values.
        """
        ...

    def delete(self) -> None:
        """
        Delete track.
        """
        ...

    @property
    def depth(self) -> int:
        """
        Track depth.

        :type: int
        """
        ...

    @property
    def envelopes(self) -> reapy.EnvelopeList:
        """
        List of envelopes on track.

        :type: EnvelopeList
        """
        ...

    @property
    def fxs(self) -> reapy.FXList:
        """
        List of FXs on track.

        :type: FXList
        """
        ...

    def get_info_string(self, param_name: str) -> str:
        ...

    def get_info_value(self, param_name: str) -> float:
        ...

    @property
    def GUID(self) -> str:
        """
        Track's GUID.

        16-byte GUID, can query or update.
        If using a _String() function, GUID is a string {xyz-...}.

        :type: str
        """
        ...

    @GUID.setter
    def GUID(self, guid_string: str) -> None:
        ...

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if track has been deleted, ID will not be valid
        anymore.

        :type: bool
        """

    @property
    def icon(self) -> str:
        """
        Track icon.

        Full filename, or relative to resource_path/data/track_icons.

        :type: str
        """
        ...

    @icon.setter
    def icon(self, filename: str) -> None:
        ...

    @property
    def index(self) -> ty.Optional[int]:
        """
        Track index in GUI (0-based).

        Will be ``None`` for master track.

        :type: int or None

        Raises
        ------
        InvalidObjectError
            When track does not exist in REAPER.
        """
        ...

    @property
    def instrument(self) -> ty.Optional[reapy.FX]:
        """
        First instrument FX on track if it exists, else None.

        :type: FX or None
        """
        ...

    @property
    def items(self) -> ty.List[reapy.Item]:
        """
        List of items on track.

        :type: list of Item
        """
        ...

    @property
    def is_muted(self) -> bool:
        """
        Whether track is muted.

        Can be manually set to change track state.
        """
        ...

    @is_muted.setter
    def is_muted(self, muted: bool) -> None:
        ...

    @property
    def is_selected(self) -> bool:
        """
        Whether track is selected.

        :type: bool
        """
        ...

    @is_selected.setter
    def is_selected(self, selected: bool) -> None:
        """
        Select or unselect track.

        Parameters
        ----------
        selected : bool
            Whether to select or unselect track.
        """
        ...

    @property
    def is_solo(self) -> bool:
        """
        Whether track is solo.

        Can be manually set to change track state.
        """
        ...

    @is_solo.setter
    def is_solo(self, solo: bool) -> None:
        ...

    def make_only_selected_track(self) -> None:
        """
        Make track the only selected track in parent project.
        """
        ...

    def midi_hash(self, notes_only: bool = False) -> str:
        """Get hash of MIDI-data to compare with later.

        Parameters
        ----------
        notes_only : bool, (False by default)
            count only notes if True

        Returns
        -------
        str
        """
        ...

    @property
    def midi_note_names(self) -> ty.List[str]:
        ...

    def mute(self) -> None:
        """Mute track (do nothing if track is already muted)."""
        ...

    @property
    def n_envelopes(self) -> int:
        """
        Number of envelopes on track.

        :type: int
        """
        ...

    @property
    def n_fxs(self) -> int:
        """
        Number of FXs on track.

        :type: int
        """
        ...

    @property
    def n_hardware_sends(self) -> int:
        """
        Number of hardware sends on track.

        :type: int
        """
        ...

    @property
    def n_items(self) -> int:
        """
        Number of items on track.

        :type: int
        """
        ...

    @property
    def n_receives(self) -> int:
        ...

    @property
    def n_sends(self) -> int:
        ...

    @property
    def name(self) -> str:
        """
        Track name.

        Name is "MASTER" for master track, "Track N" if track has no
        name.

        :type: str
            Track name .
        """
        ...

    @name.setter
    def name(self, track_name: str) -> None:
        ...

    @property
    def parent_track(self) -> ty.Optional[reapy.Track]:
        """
        Parent track, or None if track has none.

        :type: Track or NoneType
        """
        ...

    @property
    def project(self) -> reapy.Project:
        """
        Track parent project.

        :type: Project
        """
        ...

    @property
    def receives(self) -> ty.List[reapy.Send]: ...

    def select(self) -> None:
        """
        Select track.
        """
        ...

    @property
    def sends(self) -> ty.List[reapy.Send]:
        ...

    def set_info_string(self, param_name: str, param_string: str) -> None:
        ...

    def set_info_value(self, param_name: str, param_value: float) -> None: ...

    def solo(self) -> None:
        """Solo track (do nothing if track is already solo)."""
        ...

    def toggle_mute(self) -> None:
        """Toggle mute on track."""
        ...

    def toggle_solo(self) -> None:
        """Toggle solo on track."""
        ...

    def unmute(self) -> None:
        """Unmute track (do nothing if track is not muted)."""
        ...

    def unselect(self) -> None:
        """
        Unselect track.
        """
        ...

    def unsolo(self) -> None:
        """Unsolo track (do nothing if track is not solo)."""
        ...

    @property
    def visible_fx(self) -> reapy.FX:
        """
        Visible FX in FX chain if any, else None.

        :type: FX or NoneType
        """
        ...


class TrackList(ReapyObjectList):
    """
    Container for a project's track list.

    Examples
    --------
    >>> tracks = project.tracks
    >>> len(tracks)
    4
    >>> tracks[0].name
    'Kick'
    >>> for track in tracks:
    ...     print(track.name)
    ...
    'Kick'
    'Snare'
    'Hi-hat'
    'Cymbal"
    """
    parent: reapy.Project

    def __init__(self, parent: reapy.Project) -> None:
        """
        Create track list.

        Parameters
        ----------
        parent : Project
            Parent project.
        """
        ...

    @ty.overload
    def __getitem__(self, key: str) -> Track:
        ...

    @ty.overload
    def __getitem__(self, key: int) -> Track:
        ...

    @ty.overload
    def __getitem__(self, key: slice) -> ty.List[Track]:
        ...

    def __delitem__(self, key: ty.Union[int, slice]) -> None:
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> ty.Iterator[reapy.Track]:
        ...

    @property
    def _args(self) -> ty.Tuple[reapy.Project]:
        ...

    def _get_items_from_slice(self, slice: slice) -> ty.List[Track]:
        ...



import reapy
from reapy.errors import OutsideREAPERError
from reapy.reascripts import activate_reapy_server

from configparser import ConfigParser
from collections import OrderedDict
import json
import os
import shutil

import typing as ty

REAPY_SERVER_PORT = 2306
WEB_INTERFACE_PORT = 2307

T1 = ty.TypeVar('T1')
T2 = ty.TypeVar('T2')


class CaseInsensitiveDict(OrderedDict[str, T1]):
    """OrderedDict with case-insensitive keys."""
    _dict: OrderedDict[str, T1]

    def __init__(self, *args: T1, **kwargs: T1) -> None:
        ...

    def __contains__(self, key: str) -> ty.Union[str, bool]:  # type: ignore
        ...

    def __getitem__(self, key: str) -> T1:
        ...

    def __setitem__(self, key: str, value: T1) -> None:
        ...


class Config(ConfigParser):
    """Parser for REAPER .ini file."""
    ini_file: str

    def __init__(self) -> None:
        ...

    def write(self) -> None:  # type:ignore
        # Backup config state before user has ever tried reapy
        ...


def create_new_web_interface(port: int) -> None:
    """
    Create a Web interface in REAPER at a specified port.

    It is added by writing a line directly in REAPER .ini file. Thus
    it will only be available on restart.

    Parameters
    ----------
    port : int
        Web interface port.
    """
    ...


def delete_web_interface(port: int) -> None:
    """
    Delete Web interface listening to a specified port.

    It is deleted by writing a line directly in REAPER .ini file. Thus
    it will only be deleted on restart.

    Parameters
    ----------
    port : int
        Web interface port.
    """
    ...


def disable_dist_api() -> None:
    """
    Disable distant API.

    Delete ``reapy`` Web interface, and remove the ReaScript
    ``reapy.reascripts.activate_reapy_server`` from the
    Actions list.
    """
    ...


def enable_dist_api() -> None:
    """
    Enable distant API.

    Create a Web interface and add the ReaScript
    ``reapy.reascripts.activate_reapy_server`` to the Actions list.
    """
    ...


def get_activate_reapy_server_path() -> str:
    """Return path to the ``activate_reapy_server`` ReaScript."""
    ...



"""Defines class Project."""

import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
from reapy.errors import RedoError, UndoError
import typing as ty
import typing_extensions as te


class Project(ReapyObject):
    """REAPER project."""
    id: str
    _filename: str

    def __init__(self,
                 id: ty.Optional[ty.Union[int, str]] = None,
                 index: int = -1) -> None:
        """
        Build project either by ID or index.

        id : None, str or int, optional
            Project identifier.
            When None (default), `index is used instead.
            An integer is interpreted as the project index in GUI.
            A string starting with '(ReaProject*)0x' is interpreted
            as a ReaScript identifier.
            Otherwise, `id` is the project name. In that case, the .rpp
            extension is optional.
        index : int, optional
            Project index in GUI (default=-1, corresponds to current
            project).
        """
        ...

    def __eq__(self, other: object) -> bool:
        ...

    @property
    def _args(self) -> ty.Tuple[str]:
        ...

    @staticmethod
    def _from_name(name: str) -> 'Project':
        """Return project with corresponding name.

        Parameters
        ----------
        name : str
            Project file name. Including the extension ('.rpp')
            is optional.

        Returns
        -------
        Project

        Raises
        ------
        NameError
            If no project with the corresponding name is open.
        """
        ...

    def _get_track_by_name(self, name: str) -> ty.Optional[reapy.Track]:
        """Return first track with matching name."""
        ...

    def add_marker(self,
                   position: float,
                   name: str = "",
                   color: ty.Union[ty.Tuple[int], int] = 0) -> reapy.Marker:
        """
        Create new marker and return its index.

        Parameters
        ----------
        position : float
            Marker position in seconds.
        name : str, optional
            Marker name.
        color : int or tuple of int, optional
            Marker color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        marker : reapy.Marker
            New marker.

        Notes
        -----
        If a marker with the same position and name already exists, no
        new marker will be created, and the existing marker index will
        be returned.
        """
        ...

    def add_region(self,
                   start: float,
                   end: float,
                   name: str = "",
                   color: ty.Union[ty.Tuple[int], int] = 0) -> reapy.Region:
        """
        Create new region and return its index.

        Parameters
        ----------
        start : float
            Region start in seconds.
        end : float
            Region end in seconds.
        name : str, optional
            Region name.
        color : int or tuple of int, optional
            Region color. Integers correspond to REAPER native colors.
            Tuple must be RGB triplets of integers between 0 and 255.

        Returns
        -------
        region : reapy.Region
            New region.
        """
        ...

    def add_track(self, index: int = 0, name: str = "") -> reapy.Track:
        """
        Add track at a specified index.

        Parameters
        ----------
        index : int
            Index at which to insert track.
        name : str, optional
            Name of created track.

        Returns
        -------
        track : Track
            New track.
        """
        ...

    @property
    def any_track_solo(self) -> bool:
        """
        Test whether any track is soloed in project.

        Returns
        -------
        any_track_solo : bool
            Whether any track is soloed in project.
        """
        ...

    def beats_to_time(self, beats: float) -> float:
        """
        Convert beats to time in seconds.

        Parameters
        ----------
        beats : float
            Time in beats

        Returns
        -------
        time : float
            Converted time in seconds.

        See also
        --------
        Project.time_to_beats
        """
        ...

    def begin_undo_block(self) -> None:
        """
        Start a new undo block.
        """
        ...

    @property
    def bpi(self) -> float:
        """
        Return project BPI (numerator of time signature).

        Returns
        -------
        bpi : float
            Numerator of time signature.
        """
        ...

    @property
    def bpm(self) -> float:
        """
        Project BPM (beats per minute).

        :type: float
        """
        ...

    @bpm.setter
    def bpm(self, bpm: float) -> None:
        """
        Set project BPM (beats per minute).

        Parameters
        ----------
        bpm : float
            Tempo in beats per minute.
        """
        ...

    @property
    def buffer_position(self) -> float:
        """
        Position of next audio block being processed in seconds.

        :type: float

        See also
        --------
        Project.play_position
            Latency-compensated actual-what-you-hear position.
        """
        ...

    def bypass_fx_on_all_tracks(self, bypass: bool = True) -> None:
        """
        Bypass or un-bypass FX on all tracks.

        Parameters
        ----------
        bypass : bool
            Whether to bypass or un-bypass FX.
        """
        ...

    @property
    def can_redo(self) -> bool:
        """
        Whether redo is possible.

        :type: bool
        """
        ...

    @property
    def can_undo(self) -> bool:
        """
        Whether undo is possible.

        :type: bool
        """
        ...

    def close(self) -> None:
        """Colse project and its correspondig tab."""
        ...

    @property
    def cursor_position(self) -> float:
        """
        Edit cursor position in seconds.

        :type: float
        """
        ...

    @cursor_position.setter
    def cursor_position(self, position: float) -> None:
        """
        Set edit cursor position.

        Parameters
        ----------
        position : float
            New edit cursor position in seconds.
        """
        ...

    def disarm_rec_on_all_tracks(self) -> None:
        """
        Disarm record on all tracks.
        """
        ...

    def end_undo_block(self, description: str = "") -> None:
        """
        End undo block.

        Parameters
        ----------
        description : str
            Undo block description.
        """
        ...

    @property
    def focused_fx(self) -> ty.Optional[reapy.FX]:
        """
        FX that has focus if any, else None.

        :type: FX or NoneType
        """
        ...

    def get_info_string(self, param_name: str) -> str:
        ...

    def get_info_value(self, param_name: str) -> float:
        ...

    @ty.overload
    def get_ext_state(self, section: str, key: str,
                      pickled: te.Literal[True]) -> object:
        """
        Return external state of project.

        Parameters
        ----------
        section : str
        key : str
        pickled: bool
            Whether data was pickled or not.

        Returns
        -------
        value : str
            If key or section does not exist an empty string is returned.
        """
        ...

    @ty.overload
    def get_ext_state(self, section: str, key: str,
                      pickled: te.Literal[False]) -> str:

        ...

    @ty.overload
    def get_ext_state(self, section: str, key: str,
                      pickled: bool = False) -> ty.Union[str, object]:

        ...

    def get_play_rate(self, position: float) -> float:
        """
        Return project play rate at a given position.

        Parameters
        ----------
        position : float
            Position in seconds.

        Returns
        -------
        play_rate : float
            Play rate at the given position.

        See also
        --------
        Project.play_rate
            Project play rate at the current position.
        """
        ...

    def get_selected_item(self, index: int) -> reapy.Item:
        """
        Return index-th selected item.

        Parameters
        ----------
        index : int
            Item index.

        Returns
        -------
        item : Item
            index-th selected item.
        """
        ...

    def get_selected_track(self, index: int) -> reapy.Track:
        """
        Return index-th selected track.

        Parameters
        ----------
        index : int
            Track index.

        Returns
        -------
        track : Track
            index-th selected track.
        """
        ...

    def glue_items(self, within_time_selection: bool = False) -> None:
        """
        Glue items (action shortcut).

        Parameters
        ----------
        within_time_selection : bool
            If True, glue items within time selection.
        """
        ...

    @property
    def has_valid_id(self) -> bool:
        """Whether ReaScript ID is still valid.

        For instance, if project has been manually closed, ID will not
        be valid anymore.

        :type: bool
        """

    @property
    def is_dirty(self) -> bool:
        """
        Whether project is dirty (i.e. needing save).

        :type: bool
        """
        ...

    @property
    def is_current_project(self) -> bool:
        """
        Whether project is current project.

        :type: bool
        """
        ...

    @property
    def is_paused(self) -> bool:
        """
        Return whether project is paused.

        :type: bool
        """
        ...

    @property
    def is_playing(self) -> bool:
        """
        Return whether project is playing.

        :type: bool
        """
        ...

    @property
    def is_recording(self) -> bool:
        """
        Return whether project is recording.

        :type: bool
        """
        ...

    @property
    def is_stopped(self) -> bool:
        """
        Return whether project is stopped.

        :type: bool
        """
        ...

    @property
    def items(self) -> ty.List[reapy.Item]:
        """
        List of items in project.

        :type: list of Item
        """
        ...

    @property
    def length(self) -> float:
        """
        Project length in seconds.

        :type: float
        """
        ...

    @property
    def last_touched_fx(self
                        ) -> ty.Tuple[ty.Optional[reapy.FX], ty.Optional[int]]:
        """
        Last touched FX and corresponding parameter index.

        :type: FX, int or NoneType, NoneType

        Notes
        -----
        Only Track FX are detected by this property. If last touched
        FX is a Take FX, this property is ``(None, None)``.

        Examples
        --------
        >>> fx, index = project.last_touched_fx
        >>> fx.name
        'VSTi: ReaSamplOmatic5000 (Cockos)'
        >>> fx.params[index].name
        "Volume"
        """
        ...

    def make_current_project(self) -> ty.ContextManager[None]:
        """
        Set project as current project.

        Can also be used as a context manager to temporarily set
        the project as current project and then go back to the original
        situation.

        Examples
        --------
        >>> p1 = reapy.Project()  # current project
        >>> p2 = reapy.Project(1)  # other project
        >>> p2.make_current_project()  # now p2 is current project
        >>> with p1.make_current_project():
        ...     do_something()  # current project is temporarily p1
        >>> # and p2 is current project again
        """
        ...

    def mark_dirty(self) -> None:
        """
        Mark project as dirty (i.e. needing save).
        """
        ...

    @property
    def markers(self) -> ty.List[reapy.Marker]:
        """
        List of project markers.

        :type: list of reapy.Marker
        """
        ...

    @property
    def master_track(self) -> reapy.Track:
        """
        Project master track.

        :type: Track
        """
        ...

    def mute_all_tracks(self, mute: bool = True) -> None:
        """
        Mute or unmute all tracks.

        Parameters
        ----------
        mute : bool, optional
            Whether to mute or unmute all tracks (default=True).

        See also
        --------
        Project.unmute_all_tracks
        """
        ...

    @property
    def n_items(self) -> int:
        """
        Number of items in project.

        :type: int
        """
        ...

    @property
    def n_markers(self) -> int:
        """
        Number of markers in project.

        :type: int
        """
        ...

    @property
    def n_regions(self) -> int:
        """
        Number of regions in project.

        :type: int
        """
        ...

    @property
    def n_selected_items(self) -> int:
        """
        Number of selected media items.

        :type: int
        """
        ...

    @property
    def n_selected_tracks(self) -> int:
        """
        Number of selected tracks in project (excluding master).

        :type: int
        """
        ...

    @property
    def n_tempo_markers(self) -> int:
        """
        Number of tempo/time signature markers in project.

        :type: int
        """
        ...

    @property
    def n_tracks(self) -> int:
        """
        Number of tracks in project.

        :type: int
        """
        ...

    @property
    def name(self) -> str:
        """
        Project name.

        :type: str
        """
        ...

    def open(self, in_new_tab:bool=False) -> None:
        """
        Open project, if it was closed by Project.close.

        Parameters
        ----------
        in_new_tab : bool, optional
            whether should be opened in new tab

        Raises
        ------
        RuntimeError
            If hasn't been closed by Project.close yet
        """
        ...

    def pause(self) -> None:
        """
        Hit pause button.
        """
        ...

    @property
    def path(self) -> str:
        """
        Project path.

        :type: str
        """
        ...

    def perform_action(self, action_id: int) -> None:
        """
        Perform action with ID `action_id` in the main Actions section.

        Parameters
        ----------
        action_id : int
            Action ID in the main Actions section.
        """
        ...

    def play(self) -> None:
        """
        Hit play button.
        """
        ...

    @property
    def play_position(self) -> float:
        """
        Latency-compensated actual-what-you-hear position in seconds.

        :type: float

        See also
        --------
        Project.buffer_position
            Position of next audio block being processed.
        """
        ...

    @property
    def play_rate(self) -> float:
        """
        Project play rate at the cursor position.

        :type: float

        See also
        --------
        Project.get_play_rate
            Return project play rate at a specified time.
        """
        ...

    def record(self) -> None:
        """Hit record button."""
        ...

    def redo(self) -> None:
        """
        Redo last action.

        Raises
        ------
        RedoError
            If impossible to redo.
        """
        ...

    @property
    def regions(self) -> ty.List[reapy.Region]:
        """
        List of project regions.

        :type: list of reapy.Region
        """
        ...

    def save(self, force_save_as: bool = False) -> None:
        """
        Save project.

        Parameters
        ----------
        force_save_as : bool
            Force using "Save as" instead of "Save".
        """
        ...

    def select(self,
               start: float,
               end: ty.Optional[float] = None,
               length: ty.Optional[float] = None) -> None:
        ...

    def select_all_items(self, selected: bool = True) -> None:
        """
        Select or unselect all items, depending on `selected`.

        Parameters
        ----------
        selected : bool
            Whether to select or unselect items.
        """
        ...

    def select_all_tracks(self) -> None:
        """Select all tracks."""
        ...

    @property
    def selected_envelope(self) -> ty.Optional[reapy.Envelope]:
        """
        Project selected envelope.

        :type: reapy.Envelope or None
        """
        ...

    @property
    def selected_items(self) -> ty.List[reapy.Item]:
        """
        List of all selected items.

        :type: list of Item

        See also
        --------
        Project.get_selected_item
            Return a specific selected item.
        """
        ...

    @property
    def selected_tracks(self) -> ty.List[reapy.Track]:
        """
        List of selected tracks (excluding master).

        :type: list of Track
        """
        ...

    @selected_tracks.setter
    def selected_tracks(self, tracks: ty.List[reapy.Track]) -> None:
        ...

    def set_info_string(self, param_name: str, param_string: str) -> None:
        """
        Parameters
        ----------
        param_name : str
            MARKER_GUID:X : get the GUID (unique ID) of the marker or region
                with index X, where X is the index passed to
                EnumProjectMarkers, not necessarily the displayed number
            RECORD_PATH :
                recording directory -- may be blank or a relative path,
                to get the effective path see GetProjectPathEx()
            RENDER_FILE : render directory
            RENDER_PATTERN : render file name (may contain wildcards)
            RENDER_FORMAT : base64-encoded sink configuration
                (see project files, etc). Callers can also pass a simple
                4-byte string (non-base64-encoded), e.g. "evaw" or "l3pm",
                to use default settings for that sink type.
            RENDER_FORMAT2 : base64-encoded secondary sink configuration.
                Callers can also pass a simple 4-byte string (non-base64-encoded),
                e.g. "evaw" or "l3pm", to use default settings for
                that sink type, or "" to disable secondary render.
                Formats available on this machine:
                "wave" "aiff" "iso " "ddp " "flac" "mp3l" "oggv" "OggS"
                "FFMP" "GIF " "LCF " "wvpk"
        param_string : str
        """
        ...

    def set_info_value(self, param_name: str, param_value: float) -> None:
        """
        Parameters
        ----------
        param_name : str
            RENDER_SETTINGS : &(1|2)=0:master mix, &1=stems+master mix,
                &2=stems only, &4=multichannel tracks to multichannel files,
                &8=use render matrix, &16=tracks with only mono media
                to mono files, &32=selected media items,
                &64=selected media items via master
            RENDER_BOUNDSFLAG : 0=custom time bounds, 1=entire project,
                2=time selection, 3=all project regions,
                4=selected media items, 5=selected project regions
            RENDER_CHANNELS : number of channels in rendered file
            RENDER_SRATE : sample rate of rendered file
                (or 0 for project sample rate)
            RENDER_STARTPOS : render start time when RENDER_BOUNDSFLAG=0
            RENDER_ENDPOS : render end time when RENDER_BOUNDSFLAG=0
            RENDER_TAILFLAG : apply render tail setting when rendering:
                &1=custom time bounds, &2=entire project, &4=time selection,
                &8=all project regions, &16=selected media items,
                &32=selected project regions
            RENDER_TAILMS : tail length in ms to render
                (only used if RENDER_BOUNDSFLAG and RENDER_TAILFLAG are set)
            RENDER_ADDTOPROJ : 1=add rendered files to project
            RENDER_DITHER : &1=dither, &2=noise shaping, &4=dither stems,
                &8=noise shaping on stems
            PROJECT_SRATE : samplerate (ignored unless PROJECT_SRATE_USE set)
            PROJECT_SRATE_USE : set to 1 if project samplerate is used
        param_value : float
        """
        ...

    @ty.overload
    def set_ext_state(self, section: str, key: str, value: str,
                      pickled: te.Literal[False]) -> int:
        """
        Set external state of project.

        Parameters
        ----------
        section : str
        key : str
        value : Union[Any, str]
            State value. Will be dumped to str using either `pickle` if
            `pickled` is `True` or `json`. Length of the dumped value
            must not be over 2**31 - 2.
        pickled : bool, optional
            Data will be pickled with the last version if True.
            If you using mypy as type checker, typing_extensions.Literal[True]
            has to be used for `pickled`.

        Raises
        ------
        ValueError
            If dumped `value` has length over 2**31 - 2.
        """
        ...

    @ty.overload
    def set_ext_state(self, section: str, key: str, value: ty.Any,
                      pickled: te.Literal[True]) -> int: ...

    @ty.overload
    def set_ext_state(self, section: str, key: str, value: str,
                      pickled: bool = False) -> int: ...

    def solo_all_tracks(self) -> None:
        """
        Solo all tracks in project.

        See also
        --------
        Project.unsolo_all_tracks
        """
        ...

    def stop(self) -> None:
        """
        Hit stop button.
        """
        ...

    @property
    def time_selection(self) -> reapy.TimeSelection:
        """
        Project time selection.

        time_selection : reapy.TimeSelection

        Can be set and deleted as follows:

        >>> project = reapy.Project()
        >>> project.time_selection = 3, 8  # seconds
        >>> del project.time_selection
        """
        ...

    @time_selection.setter
    def time_selection(self, selection: ty.Tuple[float, float]) -> None:
        """
        Set time selection bounds.

        Parameters
        ----------
        selection : (float, float)
            Start and end of new time selection in seconds.
        """
        ...

    @time_selection.deleter
    def time_selection(self) -> None:
        """
        Delete current time selection.
        """
        ...

    @property
    def time_signature(self) -> ty.Tuple[float, float]:
        """
        Project time signature.

        This does not reflect tempo envelopes but is purely what is set in the
        project settings.

        bpm : float
            Project BPM (beats per minute)
        bpi : float
            Project BPI (numerator of time signature)
        """
        ...

    def time_to_beats(self, time: float) -> float:
        """
        Convert time in seconds to beats.

        Parameters
        ----------
        time : float
            Time in seconds.

        Returns
        -------
        beats : float
            Time in beats.

        See also
        --------
        Projecr.beats_to_time
        """
        ...

    @property
    def tracks(self) -> reapy.TrackList:
        """
        List of project tracks.

        :type: TrackList
        """
        ...

    def undo(self) -> None:
        """
        Undo last action.

        Raises
        ------
        UndoError
            If impossible to undo.
        """
        ...

    def unmute_all_tracks(self) -> None:
        """
        Unmute all tracks.
        """
        ...

    def unselect_all_tracks(self) -> None:
        """Unselect all tracks."""
        ...

    def unsolo_all_tracks(self) -> None:
        """
        Unsolo all tracks in project.

        See also
        --------
        Project.solo_all_tracks
        """
        ...


class _MakeCurrentProject:
    """Context manager used by Project.make_current_project."""
    current_project: Project

    def __init__(self, project: Project) -> None:
        ...

    def __enter__(self) -> None:
        ...

    def __exit__(self, exc_type: ty.Any, exc_val: ty.Any,
                 exc_tb: ty.Any) -> None:
        ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject
import typing as ty


class Send(ReapyObject):

    """Track send.

    Attributes
    ----------
    index : int
        position on the track
    is_muted : bool
    is_phase_flipped : bool
    track_id : str
    type : str
        can be 'send', 'hardware' or 'receive'
    """

    _class_name = "Send"
    index: int
    track_id: ty.Union[str, int]
    type: str

    def __init__(self,
                 track: ty.Optional[reapy.Track] = None,
                 index: int = 0,
                 track_id: ty.Optional[ty.Union[str, int]] = None,
                 type: str = "send") -> None:
        ...

    def _get_int_type(self) -> int:
        ...

    @property
    def _kwargs(self) -> ty.Dict[str, object]:
        ...

    def delete(self) -> None:
        """
        Delete send.
        """
        ...

    @property
    def dest_track(self) -> reapy.Track:
        """
        Destination track.

        :type: Track
        """
        ...

    @reapy.inside_reaper()
    def flip_phase(self) -> None:
        """
        Toggle whether phase is flipped.
        """
        ...

    def get_info(self, param_name: str) -> float:
        """Get raw info from GetTrackSendInfo_Value.

        Parameters
        ----------
        param_name : str
            B_MUTE : bool *
            B_PHASE : bool *, true to flip phase
            B_MONO : bool *
            D_VOL : double *, 1.0 = +0dB etc
            D_PAN : double *, -1..+1
            D_PANLAW : double *,1.0=+0.0db, 0.5=-6dB, -1.0 = projdef etc
            I_SENDMODE : int *, 0=post-fader, 1=pre-fx, 2=post-fx (deprecated),
                                3=post-fx
            I_AUTOMODE : int * : automation mode (-1=use track automode,
                                0=trim/off, 1=read, 2=touch, 3=write, 4=latch)
            I_SRCCHAN : int *, index,&1024=mono, -1 for none
            I_DSTCHAN : int *, index, &1024=mono, otherwise stereo pair,
                                hwout:&512=rearoute
            I_MIDIFLAGS : int *, low 5 bits=source channel 0=all, 1-16,
                                next 5 bits=dest channel, 0=orig,
                                1-16=chanP_DESTTRACK : read only, 
                                returns MediaTrack *,
                                destination track,
                                only applies for sends/recvs
            P_SRCTRACK : read only, returns MediaTrack *,
                                source track, only applies for sends/recvs
            P_ENV:<envchunkname : read only, returns TrackEnvelope *.
                                Call with :<VOLENV, :<PANENV, etc appended.


        Returns
        -------
        Union[bool, track id(str)]
        """
        ...

    def get_info_br(self, param_name: str) -> float:

        ...

    @property
    def is_mono(self) -> bool:
        """
        Whether send is mono or stereo.

        :type: bool
        """
        ...

    @is_mono.setter
    def is_mono(self, mono: bool) -> None:
        ...

    @property
    def is_muted(self) -> bool:
        """
        Whether send is muted.

        :type: bool
        """
        ...

    @is_muted.setter
    def is_muted(self, is_muted: bool) -> None:
        """
        Mute or unmute send.

        Parameters
        ----------
        is_muted : bool
            Whether to mute or unmute send.
        """
        ...

    @property
    def is_phase_flipped(self) -> bool:
        """
        Whether send phase is flipped (i.e. signal multiplied by -1).

        :type: bool
        """
        ...

    @is_phase_flipped.setter
    def is_phase_flipped(self, flipped: bool) -> None:
        ...

    def mute(self) -> None:
        """
        Mute send.
        """
        ...

    @property
    def _midi_flags_unpacked(self) -> ty.Tuple[
        ty.Tuple[int, int], ty.Tuple[int, int]]: ...

    @_midi_flags_unpacked.setter
    def _midi_flags_unpacked(self, in_tuple: ty.Tuple[
        ty.Tuple[int, int], ty.Tuple[int, int]]): ...

    @property
    def midi_source(self) -> ty.Tuple[int, int]:
        """
        Send MIDI properties on the send track.

        Returns
        -------
        List[int bus, int channel]
        """
        ...

    @midi_source.setter
    def midi_source(self, value: ty.Tuple[int, int]) -> None: ...

    @property
    def midi_dest(self) -> ty.Tuple[int, int]:
        """
        Send MIDI properties on the send track.

        Returns
        -------
        List[int bus, int channel]
        """
        ...

    @midi_dest.setter
    def midi_dest(self, value: ty.Tuple[int, int]) -> None: ...

    @property
    def pan(self) -> float:
        """
        Send pan (from -1=left to 1=right).

        :type: float
        """
        ...

    @pan.setter
    def pan(self, pan: float) -> None:
        """
        Set send pan.

        Parameters
        ----------
        pan : float
            New pan between -1 (left) and 1 (right).
        """
        ...

    def set_info(self, param_name: str, value: float) -> None:
        ...

    def set_info_br(self, param_name: str, value: float) -> None:
        ...

    @property
    def source_track(self) -> reapy.Track:
        """
        Source track.

        :type: Track
        """
        ...

    def unmute(self) -> None:
        """
        Unmute send.
        """
        ...

    @property
    def volume(self) -> float:
        """
        Send volume.

        :type: float
        """
        ...

    @volume.setter
    def volume(self, volume: float) -> None:
        ...



from configparser import ConfigParser
from collections import OrderedDict
import json
import os
import pathlib
import random
import re
import shutil
import string
import warnings

import reapy
from reapy.errors import OutsideREAPERError
from reapy.reascripts import activate_reapy_server
from .resource_path import get_resource_path
from .shared_library import get_python_shared_library


__all__ = [
    'add_web_interface',
    'configure_reaper',
    'create_new_web_interface',
    'delete_web_interface',
    'disable_dist_api',
    'enable_dist_api',
    'enable_python',
    'REAPY_SERVER_PORT',
    'WEB_INTERFACE_PORT'
]


REAPY_SERVER_PORT = 2306
WEB_INTERFACE_PORT = 2307


class CaseInsensitiveDict(OrderedDict):

    """OrderedDict with case-insensitive keys."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dict = OrderedDict(*args, **kwargs)
        for key, value in self._dict.items():
            self._dict[key.lower()] = value

    def __contains__(self, key):
        return key.lower() in self._dict

    def __getitem__(self, key):
        return self._dict[key.lower()]

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._dict[key.lower()] = value


class Config(ConfigParser):

    """Parser for REAPER .ini file."""

    def __init__(self, ini_file):
        super().__init__(
            strict=False, delimiters="=", dict_type=CaseInsensitiveDict
        )
        self.optionxform = str
        self.ini_file = ini_file
        if not os.path.exists(ini_file):
            pathlib.Path(ini_file).touch()
        self.read(self.ini_file, encoding='utf8')

    def write(self):
        # Backup config state before user has ever tried renardo_reapy
        before_reapy_file = self.ini_file + '.before-renardo-reapy.bak'
        if not os.path.exists(before_reapy_file):
            shutil.copy(self.ini_file, before_reapy_file)
        # Backup current config
        shutil.copy(self.ini_file, self.ini_file + '.bak')
        # Write config
        with open(self.ini_file, "w", encoding='utf8') as f:
            super().write(f, False)


def add_reascript(resource_path, script_path):
    """Add ReaScript to *Actions* list in REAPER.

    Works by manually editing ``reaper-kb.ini`` configuration file.
    Only use this function at setup time to configure REAPER.
    In other cases, make use of :func:`reapy.add_reascript`.

    In case ``script_path`` is already in Actions list, its command
    name is returned but it is not added a second time.

    Parameters
    ----------
    resource_path : str
        Path to REAPER resource directory. Can be obtained with
        :func:`reapy.config.resource_path.get_resource_path`.
    script_path : str
        Path to script that will be added.

    Returns
    -------
    str
        Action name for the newly added ReaScript.

    Raises
    ------
    FileNotFoundError
        When ``script_path`` does not exist.
    ValueError
        When ``script_path`` is not a Python module.
    """
    script_path = os.path.abspath(script_path)
    if not os.path.exists(script_path):
        raise FileNotFoundError(script_path)
    if os.path.splitext(script_path)[1] != '.py':
        raise ValueError('{} is not a Python module.'.format(script_path))
    ini_file = os.path.join(resource_path, "reaper-kb.ini")
    if not os.path.exists(ini_file):
        pathlib.Path(ini_file).touch()
    # Check if ReaScript already exists
    with open(ini_file) as f:
        lines = re.findall("^SCR 4 0 .*", f.read(), re.MULTILINE)
    for line in lines:
        if line.split(" ")[-1] == script_path:
            return '"_{}"'.format(line.split(" ")[3].strip('_'))
    # If not, add it
    code = get_new_reascript_code(ini_file)
    script_name = os.path.basename(script_path)
    new_line = 'SCR 4 0 {} "Custom: {}" {}'
    with open(ini_file, "a") as f:
        f.write(new_line.format(code, script_name, script_path))
    return '"_{}"'.format(code)


def add_web_interface(resource_path, port=WEB_INTERFACE_PORT):
    """Add a REAPER Web Interface at a specified port.

    It is added by manually editing reaper.ini configuration file,
    which is loaded on startup. Thus, the added web interface will
    only be available after restarting REAPER.

    Nothing happens in case a web interface already exists at
    ``port``.

    Parameters
    ----------
    resource_path : str
        Path to REAPER resource directory. Can be obtained with
        :func:`reapy.config.resource_path.get_resource_path`.
    port : int, optional
        Web interface port. Default=``2307``.
    """
    if web_interface_exists(resource_path, port):
        return
    config = Config(os.path.join(resource_path, "reaper.ini"))
    csurf_count = int(config["reaper"].get("csurf_cnt", "0"))
    csurf_count += 1
    config["reaper"]["csurf_cnt"] = str(csurf_count)
    key = "csurf_{}".format(csurf_count - 1)
    config["reaper"][key] = "HTTP 0 {} '' 'index.html' 0 ''".format(port)
    config.write()


def configure_reaper(resource_path=None, detect_portable_install=True):
    """Configure REAPER to allow reapy connections.

    Allows to use reapy from outside REAPER.

    Configuration is done by manually editing ``reaper.ini``
    and ``reaper-kb.ini``. It consists in the following steps:
    1. Enable usage of Python for ReaScripts.
    2. Fill in path to python shared library (.dll, .dylib or .so).
    3. Add a web interface on port 2307 to listen to reapy
       connections.
    4. Add the ReaScript ``reapy.reascripts.activate_reapy_server``
       to the *Actions* list.
    5. Add the name of this action to REAPER external state.

    It is safe to call this function several times as it only edits
    configuration files when needed.

    Parameters
    ----------
    resource_path : str or None, optional
        Path to REAPER resource directory. When ``None``, defaults to
        the result of
        :func:`reapy.config.resource_path.get_resource_path`. Use it
        if you already know where REAPER resource directory is
        located at.
    detect_portable_install : bool, optional
        If ``True``, this function will look for a currently running
        REAPER process and detect whether it is a portable install.
        If ``False``, configuration files will be looked for in the
        default locations only, which may result in a
        ``FileNotFoundError`` if no global REAPER install exists.
        Default=``True``.

    Raises
    ------
    RuntimeError
        When ``detect_portable_install=True`` and zero or more than one
        REAPER instances are currently running.
    FileNotFoundError
        When ``detect_portable_install=False`` but no global
        configuration file can be found (which means REAPER has only
        been installed as portable.)
    """
    if resource_path is None:
        resource_path = get_resource_path(detect_portable_install)
    enable_python(resource_path)
    add_web_interface(resource_path)
    action = add_reascript(resource_path, get_activate_reapy_server_path())
    set_ext_state("renardo_reapy", "activate_reapy_server", action, resource_path)


def create_new_web_interface(port):
    """Create a Web interface in REAPER at a specified port.

    .. deprecated:: 0.8.0
          ``create_new_web_interface`` will be removed in renardo_reapy 1.0.0.
          Use :func:`renardo_reapy.config.add_web_interface` that works from
          outside REAPER.

    It is added by writing a line directly in REAPER .ini file. Thus
    it will only be available on restart.

    Parameters
    ----------
    port : int
        Web interface port.
    """
    msg = (
        "Function create_new_web_interface is deprecated since 0.8.0. "
        "Use renardo_reapy.config.add_web_interface instead."
    )
    warnings.warn(FutureWarning(msg))
    config = Config(reapy.get_ini_file())
    csurf_count = int(config["reaper"].get("csurf_cnt", "0"))
    csurf_count += 1
    config["reaper"]["csurf_cnt"] = str(csurf_count)
    key = "csurf_{}".format(csurf_count - 1)
    config["reaper"][key] = "HTTP 0 {} '' 'index.html' 0 ''".format(port)
    config.write()


def delete_web_interface(resource_path, port=WEB_INTERFACE_PORT):
    """Delete a REAPER Web Interface at a specified port.

    It is deleted by manually editing reaper.ini configuration file,
    which is loaded on startup. Thus, the web interface stay alive
    until REAPER is closed.

    Parameters
    ----------
    resource_path : str
        Path to REAPER resource directory. Can be obtained with
        :func:`reapy.config.resource_path.get_resource_path`.
    port : int, optional
        Web interface port. Default=``2307``.
    """
    config = Config(os.path.join(resource_path, "reaper.ini"))
    # Get number of enabled control surfaces
    csurf_count = int(config["reaper"]["csurf_cnt"])
    # Find the one describing the web interface
    for i in range(csurf_count):
        string = config["reaper"]["csurf_{}".format(i)]
        if string.startswith("HTTP"):  # It's a web interface
            if string.split(" ")[2] == str(port):  # It's the one
                webi_index = i
    if 'webi_index' in locals():  # Means we found it
        # Remove the line...
        del config["reaper"]["csurf_{}".format(webi_index)]
        # ...and move the following lines one step above
        for i in range(webi_index, csurf_count - 1):
            next_line = config["reaper"]["csurf_{}".format(i + 1)]
            config["reaper"]["csurf_{}".format(i)] = next_line
        # Update number of control surfaces
        config["reaper"]["csurf_cnt"] = str(csurf_count - 1)
        # And write it out
        config.write()


def disable_dist_api():
    """
    Disable distant API.

    Delete ``renardo_reapy`` Web interface, and remove the ReaScript
    ``renardo_reapy.reascripts.activate_reapy_server`` from the
    Actions list.
    """
    if not reapy.is_inside_reaper():
        raise OutsideREAPERError
    delete_web_interface(reapy.get_resource_path(), WEB_INTERFACE_PORT)
    reascript_path = get_activate_reapy_server_path()
    reapy.remove_reascript(reascript_path)
    message = (
        "renardo_reapy will be disabled as soon as you restart REAPER."
    )
    reapy.show_message_box(message)


def enable_dist_api():
    """Enable distant API.

    .. deprecated:: 0.8.0
          ``enable_dist_api`` will be removed in renardo_reapy 1.0.0.
          Use :func:`renardo_reapy.config.configure_reaper` that works
          even from outside REAPER.

    Create a Web interface and add the ReaScript
    ``renardo_reapy.reascripts.activate_reapy_server`` to the Actions list.
    """
    msg = (
        "Function enable_dist_api is deprecated since 0.8.0. "
        "Use renardo_reapy.config.configure_reaper instead."
    )
    warnings.warn(FutureWarning(msg))
    if not reapy.is_inside_reaper():
        raise OutsideREAPERError
    create_new_web_interface(WEB_INTERFACE_PORT)
    reascript_path = get_activate_reapy_server_path()
    action_id = reapy.add_reascript(reascript_path)
    command_name = json.dumps(reapy.get_command_name(action_id))
    section, key, value = "renardo_reapy", "activate_reapy_server", command_name
    reapy.set_ext_state(section, key, value, persist=True)
    message = (
        "renardo_reapy successfully enabled!\n\nPlease restart REAPER.\n\nYou will "
        "then be able to import renardo_reapy from the outside."
    )
    reapy.show_message_box(message)


def enable_python(resource_path):
    shared_library = get_python_shared_library()
    config = Config(os.path.join(resource_path, "reaper.ini"))
    config["reaper"]["reascript"] = "1"
    config["reaper"]["pythonlibpath64"] = os.path.dirname(shared_library)
    config["reaper"]["pythonlibdll64"] = os.path.basename(shared_library)
    config.write()


def get_activate_reapy_server_path():
    """Return path to the ``activate_reapy_server`` ReaScript."""
    script_path = os.path.abspath(activate_reapy_server.__file__)
    if script_path.endswith(('.pyc', '.pyw')):
        script_path = script_path[:-1]
    return script_path


def get_new_reascript_code(ini_file):
    """Return new ReaScript code for reaper-kb.ini.

    Parameters
    ----------
    ini_file : str
        Path to ``reaper-kb.ini`` configuration file.

    Returns
    -------
    code : str
        ReaScript code.
    """
    def get_random_code():
        chars = string.ascii_letters + string.digits
        return "".join(random.choice(chars) for _ in range(40))
    with open(ini_file) as f:
        content = f.read()
    code = get_random_code()
    while code in content:
        code = get_random_code()
    return "RS" + code


def set_ext_state(section, key, value, resource_path):
    """Update REAPER external state.

    Works by manually editing ``reaper-extstate.ini`` configuration file.
    Only use this function at setup time to configure REAPER.
    In other cases, make use of :func:`reapy.set_ext_state`.

    Parameters
    ----------
    section : str
        External state section.
    key : str
        External state key in ``section``.
    value : str
        External state value for ``key`` in ``section``.
    resource_path : str
        Path to REAPER resource directory. Can be obtained with
        :func:`reapy.config.resource_path.get_resource_path`.

    Returns
    -------
    str
        Action name for the newly added ReaScript.
    """
    config = Config(os.path.join(resource_path, 'reaper-extstate.ini'))
    if section not in config.sections():
        config.add_section(section)
    config[section][key] = value
    config.write()


def web_interface_exists(resource_path, port=WEB_INTERFACE_PORT):
    """Return whether a REAPER Web Interface exists at a given port.

    Parameters
    ----------
    resource_path : str
        Path to REAPER resource directory. Can be obtained with
        :func:`reapy.config.resource_path.get_resource_path`.
    port : int, optional
        Web interface port. Default=``2307``.

    Returns
    -------
    bool
        Whether a REAPER Web Interface exists at ``port``.
    """
    config = Config(os.path.join(resource_path, "reaper.ini"))
    csurf_count = int(config["reaper"].get("csurf_cnt", "0"))
    for i in range(csurf_count):
        string = config["reaper"]["csurf_{}".format(i)]
        if string.startswith("HTTP"):  # It's a web interface
            if string.split(" ")[2] == str(port):  # It's the one
                return True
    return False



"""Encode and decode ``reapy`` objects as JSON."""

import importlib
import inspect
import json
import operator
import sys
import types
import typing as ty

T1 = ty.TypeVar('T1')


class ClassCache(ty.Dict[str, T1]):

    _core: ty.Optional[types.ModuleType] = None

    def __missing__(self, key: str) -> T1:
        ...


_CLASS_CACHE: ClassCache[object] = ClassCache()


class ReapyEncoder(json.JSONEncoder):
    def default(self, x: ty.Any) -> ty.Any:
        ...


def loads(s: ty.Union[ty.Text, bytes]) -> ty.Any:
    ...


def dumps(x: ty.Any) -> str:
    ...


def object_hook(x: ty.Any) -> ty.Any:
    ...



import reapy
from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class TimeSelection(ReapyObject):

    _class_name = "TimeSelection"

    def __init__(self, parent_project=None, parent_project_id=None):
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be "
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project_id = parent_project_id

    def _get_infos(self):
        """
        Return infos as returned by RPR.GetSet_LoopTimeRange2.

        Returns
        -------
        infos : tuple
            Time selection infos.
        """
        infos = RPR.GetSet_LoopTimeRange2(
            self.project_id, False, False, 0, 0, False
        )
        return infos

    def _set_start_end(self, start=None, end=None):
        infos = list(RPR.GetSet_LoopTimeRange2(
            self.project_id, False, False, 0, 0, False
        ))
        if start is None:
            start = infos[3]
        if end is None:
            end = infos[4]
        infos[1], infos[3], infos[4] = True, start, end
        RPR.GetSet_LoopTimeRange2(*infos)

    @property
    def _kwargs(self):
        return {"parent_project_id": self.project_id}

    @reapy.inside_reaper()
    @property
    def end(self):
        """
        Return time selection end in seconds.

        Returns
        -------
        end : float
            Time selection end in seconds.
        """
        infos = self._get_infos()
        end = infos[4]
        return end

    @end.setter
    def end(self, end):
        """
        Set time selection end.

        Parameters
        ----------
        end : float
            Time selection end in seconds.
        """
        self._set_start_end(end=end)

    @property
    def is_looping(self):
        """
        Return whether looping is enabled.

        Returns
        -------
        looping : bool
            Whether looping is enabled.
        """
        is_looping = bool(RPR.GetSetRepeatEx(self.project_id, -1))
        return is_looping

    @is_looping.setter
    def is_looping(self, is_looping):
        """
        Sets whether time selection should loop.

        Parameters
        ----------
        looping : bool
            Whether time selection should loop.
        """
        if is_looping:
            self.loop()
        else:
            self.unloop()

    @reapy.inside_reaper()
    @property
    def length(self):
        """
        Return time selection length in seconds.

        Returns
        -------
        length : float
            Time selection length in seconds.
        """
        infos = self._get_infos()
        start, end = infos[3:5]
        length = end - start
        return length

    @length.setter
    def length(self, length):
        """
        Set time selection length (by moving its end).

        Parameters
        ----------
        length : float
            Time selection length in seconds.
        """
        infos = list(RPR.GetSet_LoopTimeRange2(
            self.project_id, False, False, 0, 0, False
        ))
        infos[1], infos[4] = True, infos[3] + length
        RPR.GetSet_LoopTimeRange2(*infos)

    def loop(self):
        """
        Enable time selection looping.

        See also
        --------
        TimeSelection.is_looping
        TimeSelection.unloop
        """
        RPR.GetSetRepeatEx(self.project_id, 1)

    @reapy.inside_reaper()
    @property
    def start(self):
        """
        Return time selection start in seconds.

        Returns
        -------
        start : float
            Time selection start in seconds.
        """
        infos = self._get_infos()
        start = infos[3]
        return start

    @start.setter
    def start(self, start):
        """
        Set time selection start.

        Parameters
        ----------
        start : float
            New time selection start.
        """
        self._set_start_end(start)

    def shift(self, direction=""):
        """
        Shift time selection.

        Parameters
        ----------
        direction : {"right", "left"}
            Direction to which time selection will be shifted. Nothing
            happens if direction is neither "right" nor "left". Note
            that the shift size depends on whether snap is enabled
            and of the zoom level.
        """
        if direction == "right":
            RPR.Loop_OnArrow(self.project_id, 1)
        elif direction == "left":
            RPR.Loop_OnArrow(self.project_id, -1)

    def unloop(self):
        """
        Disable time selection looping.

        See also
        --------
        TimeSelection.is_looping
        TimeSelection.loop
        """
        RPR.GetSetRepeatEx(self.project_id, 0)



"""Fix several ReaScript API bugs.

All fixes will be applied to `reapy.reascript_api` by reapy during
the import process. Thus, this module is only intended to be used
internally and should not be directly used by end-users.
"""

import ctypes as ct
import re

import reapy
from reapy.reascript_api import _RPR


MAX_STRBUF = 4 * 1024 * 1024


def packp(t, v):
    m = re.match(r"^\((\w+\*|HWND)\)0x([0-9A-F]+)$", str(v))
    if m is not None:
        (_t, _v) = m.groups()
        if (_t == t or t == "void*"):
            a = int(_v[:8], 16)
            b = int(_v[8:], 16)
            p = ct.c_uint64((a << 32) | b).value
            return p
    return 0


_RPR.rpr_packp = packp


def packs_l(v, encoding="latin-1", size=MAX_STRBUF):
    return ct.create_string_buffer(str(v).encode(encoding), size)


def unpacks_l(v,  encoding="latin-1", want_raw=False):
    s = v.value if not want_raw else v.raw
    return str(s.decode(encoding))


def MIDI_GetEvt(take, evtidx, selectedOut, mutedOut, ppqposOut, msg, msg_sz):
    address = _RPR._ft["MIDI_GetEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_char_p, ct.c_void_p
    )(address)
    c_selected = ct.c_byte(selectedOut)
    c_muted = ct.c_byte(mutedOut)
    c_ppq_position = ct.c_double(ppqposOut)
    c_msg = packs_l(msg, size=msg_sz)
    c_msg_length = ct.c_int(msg_sz)
    success = f(
        _RPR.rpr_packp("MediaItem_Take*", take), ct.c_int(evtidx),
        ct.byref(c_selected), ct.byref(c_muted), ct.byref(c_ppq_position),
        c_msg, ct.byref(c_msg_length)
    )
    msg_length = c_msg_length.value
    msg = unpacks_l(c_msg, want_raw=True)[:msg_length]
    return (
        success, take, evtidx, c_selected.value, c_muted.value,
        c_ppq_position.value, msg, msg_length
    )


def MIDI_GetAllEvts(take, bufNeedBig, bufNeedBig_sz):
    address = _RPR._ft["MIDI_GetAllEvts"]
    f = ct.CFUNCTYPE(ct.c_byte, ct.c_uint64, ct.c_char_p, ct.c_void_p)(address)
    c_msg = packs_l(bufNeedBig, size=bufNeedBig_sz)
    c_msg_length = ct.c_int(bufNeedBig_sz)
    success = f(
        _RPR.rpr_packp("MediaItem_Take*", take), c_msg, ct.byref(c_msg_length)
    )
    msg_length = c_msg_length.value
    msg = unpacks_l(c_msg, want_raw=True)[:msg_length]
    return success, take, msg, msg_length


def MIDI_GetHash(p0, p1, p2, p3):
    a = _RPR._ft["MIDI_GetHash"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_byte,  ct.c_char_p, ct.c_int
    )(a)
    t = (
        _RPR.rpr_packp("MediaItem_Take*", p0), ct.c_byte(p1), packs_l(p2),
        ct.c_int(p3)
    )
    r = f(*t)
    return r, p0, p1, unpacks_l(t[2]), p3


def MIDI_GetTextSysexEvt(p0, p1, p2, p3, p4, p5, p6, p7):
    address = _RPR._ft["MIDI_GetTextSysexEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_void_p, ct.c_char_p, ct.c_void_p
    )(address)
    c_selected = ct.c_byte(p2)
    c_muted = ct.c_byte(p3)
    c_ppq_position = ct.c_double(p4)
    c_type = ct.c_int(p5)
    c_msg = packs_l(p6, size=p7)
    c_msg_length = ct.c_int(p7)
    success = f(
        _RPR.rpr_packp("MediaItem_Take*", p0), ct.c_int(p1),
        ct.byref(c_selected), ct.byref(c_muted), ct.byref(c_ppq_position),
        ct.byref(c_type), c_msg, ct.byref(c_msg_length)
    )
    msg_length = c_msg_length.value
    msg = unpacks_l(c_msg, want_raw=True)[:msg_length]
    return (
        success, p0, p1, c_selected.value, c_muted.value, c_ppq_position.value,
        c_type.value, msg, msg_length
    )


def MIDI_GetTrackHash(p0, p1, p2, p3):
    a = _RPR._ft["MIDI_GetTrackHash"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_byte, ct.c_char_p, ct.c_int
    )(a)
    t = (
        _RPR.rpr_packp("MediaTrack*", p0), ct.c_byte(p1), packs_l(p2),
        ct.c_int(p3)
    )
    r = f(*t)
    return r, p0, p1, unpacks_l(t[2]), p3


def MIDI_InsertEvt(take, selected, muted, ppqpos, bytestr, bytestr_sz):
    a = _RPR._ft["MIDI_InsertEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_byte, ct.c_byte, ct.c_double, ct.c_char_p,
        ct.c_int
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_byte(selected),
        ct.c_byte(muted),
        ct.c_double(ppqpos),
        packs_l(bytestr),
        ct.c_int(bytestr_sz)
    )


def MIDI_InsertTextSysexEvt(
    take, selected, muted, ppqpos, type_, bytestr, bytestr_sz
):
    a = _RPR._ft["MIDI_InsertTextSysexEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_byte, ct.c_byte, ct.c_double,
        ct.c_int, ct.c_char_p, ct.c_int
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_byte(selected),
        ct.c_byte(muted),
        ct.c_double(ppqpos),
        ct.c_int(type_),
        packs_l(bytestr),
        ct.c_int(bytestr_sz)
    )


def MIDI_SetAllEvts(take, buf, buf_sz):
    a = _RPR._ft["MIDI_SetAllEvts"]
    f = ct.CFUNCTYPE(ct.c_byte, ct.c_uint64, ct.c_char_p, ct.c_int)(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        packs_l(buf, size=buf_sz),
        ct.c_int(buf_sz)
    )


def MIDI_SetCC(
    take, ccidx, selected, muted, ppqpos, chan_msg, channel, msg2, msg3, sort
):
    a = _RPR._ft["MIDI_SetCC"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p,
        ct.c_void_p
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_int(ccidx),
        None if selected is None else ct.byref(ct.c_byte(selected)),
        None if muted is None else ct.byref(ct.c_byte(muted)),
        None if ppqpos is None else ct.byref(ct.c_double(ppqpos)),
        None if chan_msg is None else ct.byref(ct.c_int(chan_msg)),
        None if channel is None else ct.byref(ct.c_int(channel)),
        None if msg2 is None else ct.byref(ct.c_int(msg2)),
        None if msg3 is None else ct.byref(ct.c_int(msg3)),
        None if sort is None else ct.c_byte(sort)
    )


def MIDI_SetCCShape(take, ccidx, shape, beiz_tens, no_sort):
    a = _RPR._ft["MIDI_SetCCShape"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_int, ct.c_double, ct.c_void_p
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_int(ccidx),
        ct.c_int(shape),
        ct.c_double(beiz_tens),
        None if no_sort is None else ct.byref(ct.c_byte(no_sort))
    )


def MIDI_SetEvt(take, evt_idx, selected, muted, ppqpos, msg, msg_sz, no_sort):
    a = _RPR._ft["MIDI_SetEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_char_p, ct.c_int, ct.c_void_p
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_int(evt_idx),
        None if selected is None else ct.byref(ct.c_byte(selected)),
        None if muted is None else ct.byref(ct.c_byte(muted)),
        None if ppqpos is None else ct.byref(ct.c_double(ppqpos)),
        None if msg is None else packs_l(msg),
        ct.c_int(msg_sz if msg_sz is not None else 0),
        None if no_sort is None else ct.byref(ct.c_byte(no_sort))
    )


def MIDI_SetNote(
    take, idx, selected, muted, startppqpos, endppqpos, chan, pitch, vel,
    noSort
):
    a = _RPR._ft["MIDI_SetNote"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_void_p,
        ct.c_void_p
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_int(idx),
        None if selected is None else ct.byref(ct.c_byte(selected)),
        None if muted is None else ct.byref(ct.c_byte(muted)),
        None if startppqpos is None else ct.byref(ct.c_double(startppqpos)),
        None if endppqpos is None else ct.byref(ct.c_double(endppqpos)),
        None if chan is None else ct.byref(ct.c_int(chan)),
        None if pitch is None else ct.byref(ct.c_int(pitch)),
        None if vel is None else ct.byref(ct.c_int(vel)),
        None if noSort is None else ct.byref(ct.c_byte(noSort))
    )


def MIDI_SetTextSysexEvt(
    take, idx, selected, muted, ppqpos, type_, msg, msg_sz, noSort
):
    a = _RPR._ft["MIDI_SetTextSysexEvt"]
    f = ct.CFUNCTYPE(
        ct.c_byte, ct.c_uint64, ct.c_int, ct.c_void_p, ct.c_void_p,
        ct.c_void_p, ct.c_void_p, ct.c_char_p, ct.c_int, ct.c_void_p
    )(a)
    return f(
        _RPR.rpr_packp("MediaItem_Take*", take),
        ct.c_int(idx),
        None if selected is None else ct.byref(ct.c_byte(selected)),
        None if muted is None else ct.byref(ct.c_byte(muted)),
        None if ppqpos is None else ct.byref(ct.c_double(ppqpos)),
        None if type_ is None else ct.byref(ct.c_int(type_)),
        None if msg is None else packs_l(msg, size=msg_sz),
        ct.c_int(0) if msg_sz is None else ct.c_int(msg_sz),
        None if noSort is None else ct.byref(ct.c_byte(noSort))
    )


@reapy.inside_reaper()
def ValidatePtr2(p0, p1, p2):
    a = _RPR._ft["ValidatePtr2"]
    f = ct.CFUNCTYPE(ct.c_byte, ct.c_uint64, ct.c_uint64, ct.c_char_p)(a)
    project = _RPR.rpr_packp("ReaProject*", p0)
    pointer = ct.c_uint64(p1)
    name = _RPR.rpr_packsc(p2)
    return f(project, pointer, name)



import typing as ty

from reapy import reascript_api as RPR
from reapy.core import ReapyObject


class Source(ReapyObject):
    id: str

    def __init__(self, id: str) -> None:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    @property
    def _args(self) -> ty.Tuple[str]:
        ...

    def delete(self) -> None:
        """
        Delete source. Be sure that no references to source remains.
        """
        ...

    @property
    def filename(self) -> str:
        """
        Return source file name.

        Returns
        -------
        filename : str
            Source file name.
        """
        ...

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if source has been deleted, ID will not be valid
        anymore.

        :type: bool
        """

    def length(self, unit: str = "seconds") -> float:
        """
        Return source length in `unit`.

        Parameters
        ----------
        unit : {"beats", "seconds"}

        Returns
        -------
        length : float
            Source length in `unit`.
        """
        ...

    @property
    def n_channels(self) -> int:
        """
        Return number of channels in source media.

        Returns
        -------
        n_channels : int
            Number of channels in source media.
        """
        ...

    @property
    def sample_rate(self) -> int:
        """
        Return source sample rate.

        Returns
        -------
        sample_rate : int
            Source sample rate.
        """
        ...

    @property
    def type(self) -> str:
        """
        Return source type ("WAV, "MIDI", etc.).

        Returns
        -------
        type : str
            Source type.
        """
        ...



import reapy
import reapy.reascript_api as RPR
from reapy.core import ReapyObject, ReapyObjectList
import typing as ty
import typing_extensions as te

T = ty.TypeVar('T')


class MIDIEvent(ReapyObject):
    """Abstract class for MIDI events."""
    parent: reapy.Take
    index: int

    def __init__(self, parent: reapy.Take, index: int):
        """
        Create event.

        Parameters
        ----------
        parent : Take
            Take to which the event belongs.
        index : int
            Event index. It is specific to the event type, which means
            a Note and a CC can have the same index.
        """
        ...

    @property
    def _args(self) -> ty.Tuple[reapy.Take, int]:
        ...


class MIDIEventList(ReapyObjectList, ty.Generic[T]):
    parent: reapy.Take

    def __init__(self, parent: reapy.Take) -> None:
        """
        Create event list.

        Parameters
        ----------
        parent : Take
            Take to which the event list belongs.
        """
        ...

    def __getitem__(self, key: int) -> T:
        ...

    def __len__(self) -> int:
        ...

    @property
    def _args(self) -> ty.Tuple[reapy.Take]:
        ...

    @property
    def _elements_class(self) -> ty.Type[T]:
        ...

    @property
    def _n_elements(self) -> str:
        ...


CC_INFOS_T = te.TypedDict(
    'CC_INFOS_T', {
        "selected": bool,
        "muted": bool,
        "position": float,
        "channel": int,
        "channel_message": int,
        "messages": ty.Tuple[int, int]
    })


class CC(MIDIEvent):
    """MIDI CC event."""
    @property
    def channel(self) -> int:
        """
        CC channel between 0 and 15.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...

    @property
    def channel_message(self) -> int:
        """
        CC channel message.

        :type: int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...

    @property
    def infos(self) -> CC_INFOS_T:
        """
        Return infos about CC.

        Keys are {"selected", "muted", "position", "channel",
        "channel_message", "messages"}.

        :type: dict
        """
        ...

    @property
    def messages(self) -> ty.Tuple[int, int]:
        """
        CC messages.

        :type: 2-list of int

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...

    @property
    def muted(self) -> bool:
        """
        Whether CC is muted.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...

    @property
    def position(self) -> float:
        """
        CC position in seconds.

        :type: float

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...

    @property
    def selected(self) -> bool:
        """
        Whether CC is selected.

        :type: bool

        See also
        --------
        CC.infos
            For maximum efficiency when querying several properties of
            a CC.
        """
        ...


class CCList(MIDIEventList[CC]):

    _elements_class: ty.Type[CC]
    _n_elements: str


NOTE_INFOS_T = te.TypedDict(
    'NOTE_INFOS_T', {
        "selected": bool,
        "muted": bool,
        "position": float,
        "channel": int,
        "pitch": int,
        "velocity": int
    })


class Note(MIDIEvent):
    """MIDI note."""
    @property
    def channel(self) -> int:
        """
        Note channel between 0 and 15.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def end(self) -> float:
        """
        Note end in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def infos(self) -> NOTE_INFOS_T:
        """
        Return infos about note.

        Keys are {"selected", "muted", "start", "end", "channel",
        "pitch", "velocity"}.

        :type: dict
        """
        ...

    @property
    def muted(self) -> bool:
        """
        Whether note is muted.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def pitch(self) -> int:
        """
        Note pitch between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def selected(self) -> bool:
        """
        Whether note is selected.

        :type: bool

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def start(self) -> float:
        """
        Note start in seconds.

        :type: float

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...

    @property
    def velocity(self) -> int:
        """
        Note velocity between 0 and 127.

        :type: int

        See also
        --------
        Note.infos
            For maximum efficiency when querying several properties of
            a Note.
        """
        ...


class NoteList(MIDIEventList[Note]):

    _elements_class: ty.Type[Note]
    _n_elements: str



import reapy.reascript_api as RPR
from .window import Window


class ToolTip(Window):
    """Tooltip window."""
    id: bytes
    _x: int
    _y: int
    _topmost: bool
    _show: bool

    def __init__(self,
                 message: str = " ",
                 x: int = 0,
                 y: int = 0,
                 topmost: bool = True,
                 show: bool = True) -> None:
        """Initialize tooltip.

        Parameters
        ----------
        message : str, optional
            ToolTip message (default=" "). Note that tooltips with
            empty messages are always hidden.
        x : int, optional
            x position (default=0).
        y : int, optional
            y position (default=0).
        topmost : bool, optional
            Whether tooltip should be displayed on top of all other
            windows (default=True).
        show : bool, optional
            Whether to show tooltip on initialization (default=True).
        """
        ...

    def hide(self) -> None:
        """Hide tooltip."""
        ...

    @property
    def message(self) -> str:
        """
        Tooltip message.

        Note that tooltips with empty messages are always hidden.

        :type: str
        """
        ...

    @message.setter
    def message(self, message: str) -> None:
        ...

    def refresh(self) -> None:
        ...

    def show(self) -> None:
        """Show tooltip."""
        ...

    @property
    def topmost(self) -> bool:
        """
        Whether tooltip is displayed on top of all other windows.

        :type: bool
        """
        ...

    @topmost.setter
    def topmost(self, topmost: bool) -> None:
        ...

    @property
    def x(self) -> int:
        """
        x position.

        :type: int"""
        ...

    @x.setter
    def x(self, x: int) -> None:
        ...

    @property
    def y(self) -> int:
        """y position.

        :type: int
        """
        ...

    @y.setter
    def y(self, y: int) -> None:
        ...



"""Define custom errors."""

import typing as ty


class DisabledDistAPIError(Exception):
    ...


class DisabledDistAPIWarning(Warning):
    ...


class DisconnectedClientError(Exception):
    ...


class DistError(Exception):
    ...


class ExtensionNotFoundError(Exception):
    def __init__(self, extension: str, url: str): ...


class InsideREAPERError(Exception):
    ...


class InvalidObjectError(Exception):

    """Raised when an object with invalid ID has tried to access REAPER.

    Common causes of this error are closing REAPER or deleting the
    object referred to by the aforementioned ID.

    The object that caused this error is available as its ``object``
    attribute.

    Parameters
    ----------
    object : ReapyObject
        Object that caused the error.

    Notes
    -----
    Most reapy objects have a ``has_valid_id()`` method that allows
    to check for its validity.
    """

    ...


class OutsideREAPERError(Exception):
    ...


class RedoError(Exception):
    ...


class UndefinedEnvelopeError(Exception):
    def __init__(self, index: ty.Optional[int], name: ty.Optional[str],
                 chunk_name: ty.Optional[str]) -> None:
        ...


class UndefinedExtStateError(Exception):
    def __init__(self, key: object) -> None:
        ...


class UndefinedFXParamError(Exception):
    def __init__(self, fx_name: str, name: str) -> None:
        ...


class UndoError(Exception):
    def __init__(self) -> None:
        ...



import socket
import typing as ty
FuncType = ty.Callable[..., ty.Any]
F = ty.TypeVar('F', bound=FuncType)
T = ty.TypeVar('T')


class Socket:
    """
    Wrapped `socket` that can send and receive data of any length.
    """
    _socket: socket.socket

    def __init__(self, s: ty.Optional['Socket'] = None) -> None:
        ...

    @staticmethod
    def _non_blocking(f: F) -> F:
        """
        Modify a socket method so that it returns `None` when time
        out is reached.
        """
        ...

    def accept(self, *args: ty.Any, **kwargs: ty.Any
               ) -> ty.Tuple['Socket', ty.Union[str, ty.Tuple[str, ...]]]:
        ...

    def bind(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...

    def close(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...

    def connect(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...

    def listen(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...

    def recv(self, timeout: float = .0001) -> bytes:
        """Receive data of arbitrary length."""
        ...

    def send(self, data: bytes) -> None:
        """Send data."""
        ...

    def settimeout(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...

    def shutdown(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        ...
